@@ -12,6 +12,9 @@
 #if defined(STDC_HEADERS)
 #include <stdlib.h>
 #endif
+#ifdef HAVE_STRING_H
+#include <string.h>
+#endif
 #include  "fetchmail.h"
 #include  "socket.h"
 
@@ -56,11 +59,12 @@ static int pop2_getauth(int sock, struct query *ctl, char *buf)
 {
     int status;
 
-    strcpy(shroud, ctl->password);
+    strlcpy(shroud, ctl->password, sizeof(shroud));
     status = gen_transact(sock,
 		  "HELO %s %s",
 		  ctl->remotename, ctl->password);
     shroud[0] = '\0';
+    return status;
 }
 
 static int pop2_getrange(int sock, struct query *ctl, const char *folder, 
@@ -113,7 +117,7 @@ static int pop2_fetch(int sock, struct query *ctl, int number, int *lenp)
     return(ok);
 }
 
-static int pop2_trail(int sock, struct query *ctl, int number)
+static int pop2_trail(int sock, struct query *ctl, int number, const char *tag)
 /* send acknowledgement for message data */
 {
     return(gen_transact(sock, ctl->keep ? "ACKS" : "ACKD"));
@@ -125,16 +129,11 @@ static int pop2_logout(int sock, struct query *ctl)
     return(gen_transact(sock, "QUIT"));
 }
 
-const static struct method pop2 =
+static const struct method pop2 =
 {
     "POP2",				/* Post Office Protocol v2 */
-#if INET6_ENABLE
     "pop2",				/* standard POP2 port */
-    "pop2",				/* ssl POP2 port */
-#else /* INET6_ENABLE */
-    109,				/* standard POP2 port */
-    109,				/* ssl POP2 port - not */
-#endif /* INET6_ENABLE */
+    "pop2",				/* ssl POP2 port - not */
     FALSE,				/* this is not a tagged protocol */
     FALSE,				/* does not use message delimiter */
     pop2_ok,				/* parse command response */
