@@ -10,7 +10,7 @@
 #include  "config.h"
 #include  <stdio.h>
 #include  <string.h>
-#include  <ctype.h> /* isspace() */
+#include  <ctype.h>
 #ifdef HAVE_MEMORY_H
 #include  <memory.h>
 #endif /* HAVE_MEMORY_H */
@@ -29,26 +29,31 @@
 #ifdef HAVE_NET_SOCKET_H
 #include <net/socket.h>
 #endif
+#include <sys/socket.h>
+#include <netdb.h>
+#include "md5.h"
 
 #include "i18n.h"
 #include "socket.h"
 #include "fetchmail.h"
 
-#ifndef strstr		/* glibc-2.1 declares this as a macro */
-extern char *strstr();	/* needed on sysV68 R3V7.1. */
-#endif /* strstr */
-
-int mytimeout;		/* value of nonreponse timeout */
-int suppress_tags;	/* emit tags? */
-char shroud[PASSWORDLEN*2+3];	/* string to shroud in debug output */
-struct msgblk msgblk;
+/* global variables: please reinitialize them explicitly for proper
+ * working in daemon mode */
 
+/* session variables initialized in init_transact() */
+int suppress_tags = FALSE;	/* emit tags? */
 char tag[TAGLEN];
 static int tagnum;
 #define GENSYM	(sprintf(tag, "A%04d", ++tagnum % TAGMOD), tag)
-
-static int accept_count, reject_count;
 static struct method *protocol;
+char shroud[PASSWORDLEN*2+3];	/* string to shroud in debug output */
+
+/* session variables initialized in do_session() */
+int mytimeout;		/* value of nonreponse timeout */
+
+/* mail variables initialized in readheaders() */
+struct msgblk msgblk;
+static int accept_count, reject_count;
 
 static void map_name(const char *name, struct query *ctl, struct idlist **xmit_names)
 /* add given name to xmit_names if it matches declared localnames */
@@ -86,7 +91,7 @@ static void find_server_names(const char *hdr,
     {
 	char	*cp;
 
-	for (cp = nxtaddr(hdr);
+	for (cp = nxtaddr((const unsigned char *)hdr);
 	     cp != NULL;
 	     cp = nxtaddr(NULL))
 	{
@@ -109,11 +114,11 @@ static void find_server_names(const char *hdr,
 	    {
 		int sl = strlen(ctl->server.qvirtual);
  
-		if (!strncasecmp(cp, ctl->server.qvirtual, sl))
+		if (!strncasecmp((char *)cp, ctl->server.qvirtual, sl))
 		    cp += sl;
 	    }
 
-	    if ((atsign = strchr(cp, '@'))) {
+	    if ((atsign = strchr((char *)cp, '@'))) {
 		struct idlist	*idp;
 
 		/*
@@ -124,15 +129,15 @@ static void find_server_names(const char *hdr,
 		for (idp = ctl->server.localdomains; idp; idp = idp->next) {
 		    char	*rhs;
 
-		    rhs = atsign + (strlen(atsign) - strlen(idp->id));
+		    rhs = atsign + (strlen(atsign) - strlen((char *)idp->id));
 		    if (rhs > atsign &&
 			(rhs[-1] == '.' || rhs[-1] == '@') &&
-			strcasecmp(rhs, idp->id) == 0)
+			strcasecmp(rhs, (char *)idp->id) == 0)
 		    {
 			if (outlevel >= O_DEBUG)
 			    report(stdout, GT_("passed through %s matching %s\n"), 
 				  cp, idp->id);
-			save_str(xmit_names, cp, XMIT_ACCEPT);
+			save_str(xmit_names, (const char *)cp, XMIT_ACCEPT);
 			accept_count++;
 			goto nomap;
 		    }
@@ -206,17 +211,17 @@ static char *parse_received(struct query *ctl, char *bufp)
     {
 	if (!(ok = strstr(base, "by")))
 	    break;
-	else if (!isspace(ok[-1]) || !isspace(ok[2]))
+	else if (!isspace((unsigned char)ok[-1]) || !isspace((unsigned char)ok[2]))
 	    continue;
 	else
 	{
 	    char	*sp, *tp;
 
 	    /* extract space-delimited token after "by" */
-	    for (sp = ok + 2; isspace(*sp); sp++)
+	    for (sp = ok + 2; isspace((unsigned char)*sp); sp++)
 		continue;
 	    tp = rbuf;
-	    for (; !isspace(*sp); sp++)
+	    for (; *sp && !isspace((unsigned char)*sp); sp++)
 		RBUF_WRITE(*sp);
 	    *tp = '\0';
 
@@ -254,17 +259,17 @@ static char *parse_received(struct query *ctl, char *bufp)
 	{
 	    if (!(ok = strstr(base, "for")))
 		break;
-	    else if (!isspace(ok[-1]) || !isspace(ok[3]))
+	    else if (!isspace((unsigned char)ok[-1]) || !isspace((unsigned char)ok[3]))
 		continue;
 	    else
 	    {
 		char	*sp, *tp;
 
 		/* extract space-delimited token after "for" */
-		for (sp = ok + 3; isspace(*sp); sp++)
+		for (sp = ok + 3; isspace((unsigned char)*sp); sp++)
 		    continue;
 		tp = rbuf;
-		for (; !isspace(*sp); sp++)
+		for (; !isspace((unsigned char)*sp); sp++)
 		    RBUF_WRITE(*sp);
 		*tp = '\0';
 
@@ -280,7 +285,7 @@ static char *parse_received(struct query *ctl, char *bufp)
 	    char	*sp, *tp;
 
 	    /* char after "for" could be space or a continuation newline */
-	    for (sp = ok + 4; isspace(*sp); sp++)
+	    for (sp = ok + 4; isspace((unsigned char)*sp); sp++)
 		continue;
 	    tp = rbuf;
 	    RBUF_WRITE(':');	/* Here is the hack.  This is to be friends */
@@ -294,9 +299,9 @@ static char *parse_received(struct query *ctl, char *bufp)
 		while (*sp && *sp++ != ':')
 		    continue;
             while (*sp
-                   && (want_gt ? (*sp != '>') : !isspace(*sp))
+                   && (want_gt ? (*sp != '>') : !isspace((unsigned char)*sp))
                    && *sp != ';')
-		if (!isspace(*sp))
+		if (!isspace((unsigned char)*sp))
 		{
 		    RBUF_WRITE(*sp);
 		    sp++;
@@ -378,7 +383,9 @@ int readheaders(int sock,
     int			env_offs;
     char		*received_for, *rcv, *cp;
     static char		*delivered_to = NULL;
-    int 		n, linelen, oldlen, ch, remaining, skipcount;
+    int 		n, oldlen, ch, remaining, skipcount;
+    size_t		linelen;
+    int			delivered_to_count;
     struct idlist 	*idp;
     flag		no_local_matches = FALSE;
     flag		has_nuls;
@@ -400,70 +407,83 @@ int readheaders(int sock,
      * condition the code for sending bouncemail will actually look
      * at the freed storage and coredump...
      */
-    if (msgblk.headers)
-       free(msgblk.headers);
+    xfree(msgblk.headers);
     free_str_list(&msgblk.recipients);
-    if (delivered_to)
-	free(delivered_to);
+    xfree(delivered_to);
 
-    /* initially, no message ID */
-    if (ctl->thisid)
-	free(ctl->thisid);
-    ctl->thisid = NULL;
+    /* initially, no message digest */
+    memset(ctl->digest, '\0', sizeof(ctl->digest));
 
-    msgblk.headers = received_for = delivered_to = NULL;
+    received_for = NULL;
     from_offs = reply_to_offs = resent_from_offs = app_from_offs = 
 	sender_offs = resent_sender_offs = env_offs = -1;
     oldlen = 0;
     msgblk.msglen = 0;
     skipcount = 0;
+    delivered_to_count = 0;
     ctl->mimemsg = 0;
 
     for (remaining = fetchlen; remaining > 0 || protocol->delimited; )
     {
 	char *line, *rline;
-	int overlong = FALSE;
+	int overlong = FALSE; /* XXX FIXME: this is unused */
 
 	line = xmalloc(sizeof(buf));
 	linelen = 0;
 	line[0] = '\0';
 	do {
-	    set_timeout(mytimeout);
-	    if ((n = SockRead(sock, buf, sizeof(buf)-1)) == -1) {
+	    do {
+		char	*sp, *tp;
+
+		set_timeout(mytimeout);
+		if ((n = SockRead(sock, buf, sizeof(buf)-1)) == -1) {
+		    set_timeout(0);
+		    free(line);
+		    return(PS_SOCKET);
+		}
 		set_timeout(0);
-		free(line);
-		free(msgblk.headers);
-		msgblk.headers = NULL;
-		return(PS_SOCKET);
-	    }
-	    set_timeout(0);
+
+		/*
+		 * Smash out any NULs, they could wreak havoc later on.
+		 * Some network stacks seem to generate these at random,
+		 * especially (according to reports) at the beginning of the
+		 * first read.  NULs are illegal in RFC822 format.
+		 */
+		for (sp = tp = buf; sp < buf + n; sp++)
+		    if (*sp)
+			*tp++ = *sp;
+		*tp = '\0';
+		n = tp - buf;
+	    } while
+		  (n == 0);
 
 	    remaining -= n;
 	    linelen += n;
 	    msgblk.msglen += n;
 
-		/*
-		 * Try to gracefully handle the case, where the length of a
-		 * line exceeds MSGBUFSIZE.
-		 */
-		if ( n && buf[n-1] != '\n' ) {
-			overlong = TRUE;
-			rline = (char *) realloc(line, linelen);
-			if (rline == NULL)
-			{
-			    free (line);
-			    return(PS_IOERR);
-			}
-			line = rline;
-			memcpy(line + linelen - n, buf, n);
-			ch = ' '; /* So the next iteration starts */
-			continue;
+	    /*
+	     * Try to gracefully handle the case where the length of a
+	     * line exceeds MSGBUFSIZE.
+	     */
+	    if (n && buf[n-1] != '\n') 
+	    {
+		overlong = TRUE;
+		rline = (char *) realloc(line, linelen + 1);
+		if (rline == NULL)
+		{
+		    free (line);
+		    return(PS_IOERR);
 		}
-
+		line = rline;
+		memcpy(line + linelen - n, buf, n);
+		line[linelen] = '\0';
+		ch = ' '; /* So the next iteration starts */
+		continue;
+	    }
 
 	    /* lines may not be properly CRLF terminated; fix this for qmail */
-		/* we don't want to overflow the buffer here */
-	    if (ctl->forcecr && buf[n-1] == '\n' && (n == 1 || buf[n-2] != '\r'))
+	    /* we don't want to overflow the buffer here */
+	    if (ctl->forcecr && buf[n-1]=='\n' && (n==1 || buf[n-2]!='\r'))
 	    {
 		char * tcp;
 		rline = (char *) realloc(line, linelen + 2);
@@ -526,7 +546,7 @@ int readheaders(int sock,
 	     * line before the body! Without this check fetchmail segfaults.
 	     * With it, we treat such messages as spam and refuse them.
 	     */
-	    if (!refuse_mail && !isspace(line[0]) && !strchr(line, ':'))
+	    if (!refuse_mail && !isspace((unsigned char)line[0]) && !strchr(line, ':'))
 	    {
 		if (linelen != strlen (line))
 		    has_nuls = TRUE;
@@ -551,7 +571,7 @@ int readheaders(int sock,
 	    sizeticker += linelen;
 	    while (sizeticker >= SIZETICKER)
 	    {
-		if ((!run.use_syslog && !isafile(1)) || run.showdots)
+		if (outlevel > O_SILENT && run.showdots && !run.use_syslog)
 		{
 		    fputc('.', stdout);
 		    fflush(stdout);
@@ -587,29 +607,44 @@ int readheaders(int sock,
 	if (linelen != strlen (line))
 	    has_nuls = TRUE;
 
-	/* save the message's ID, we may use it for killing duplicates later */
-	if (MULTIDROP(ctl) && !strncasecmp(line, "Message-ID:", 11))
-	    ctl->thisid = xstrdup(line);
-
 	/*
 	 * The University of Washington IMAP server (the reference
 	 * implementation of IMAP4 written by Mark Crispin) relies
 	 * on being able to keep base-UID information in a special
 	 * message at the head of the mailbox.  This message should
 	 * neither be deleted nor forwarded.
+	 *
+	 * An example for such a message is (keep this in so people
+	 * find it when looking where the special code is to handle the
+	 * data):
+	 *
+	 *   From MAILER-DAEMON Wed Nov 23 11:38:42 2005
+	 *   Date: 23 Nov 2005 11:38:42 +0100
+	 *   From: Mail System Internal Data <MAILER-DAEMON@mail.example.org>
+	 *   Subject: DON'T DELETE THIS MESSAGE -- FOLDER INTERNAL DATA
+	 *   Message-ID: <1132742322@mail.example.org>
+	 *   X-IMAP: 1132742306 0000000001
+	 *   Status: RO
+	 *
+	 *   This text is part of the internal format of your mail folder, and is not
+	 *   a real message.  It is created automatically by the mail system software.
+	 *   If deleted, important folder data will be lost, and it will be re-created
+	 *   with the data reset to initial values.
+	 *
+	 * This message is only visible if a POP3 server that is unaware
+	 * of these UWIMAP messages is used besides UWIMAP or PINE.
+	 *
+	 * We will just check if the first message in the mailbox has an
+	 * X-IMAP: header.
 	 */
 #ifdef POP2_ENABLE
 	/*
 	 * We disable this check under POP2 because there's no way to
-	 * prevent deletion of the message.  So at least we ought to 
+	 * prevent deletion of the message.  So at least we ought to
 	 * forward it to the user so he or she will have some clue
 	 * that things have gone awry.
 	 */
-#if INET6_ENABLE
-	if (strncmp(protocol->service, "pop2", 4))
-#else /* INET6_ENABLE */
-	if (protocol->port != 109)
-#endif /* INET6_ENABLE */
+	if (servport("pop2") != servport(protocol->service))
 #endif /* POP2_ENABLE */
 	    if (num == 1 && !strncasecmp(line, "X-IMAP:", 7)) {
 		free(line);
@@ -645,17 +680,24 @@ int readheaders(int sock,
 	}
 
 	/*
-	 * We remove all Delivered-To: headers.
-	 * 
-	 * This is to avoid false mail loops messages when delivering
-	 * local messages to and from a Postfix/qmail mailserver. 
+	 * We remove all Delivered-To: headers if dropdelivered is set
+	 * - special care must be taken if Delivered-To: is also used
+	 * as envelope at the same time.
+	 *
+	 * This is to avoid false mail loops errors when delivering
+	 * local messages to and from a Postfix or qmail mailserver.
 	 */
 	if (ctl->dropdelivered && !strncasecmp(line, "Delivered-To:", 13)) 
 	{
-	    if (delivered_to)
+	    if (delivered_to ||
+	    	ctl->server.envelope == STRING_DISABLED ||
+		!ctl->server.envelope ||
+		strcasecmp(ctl->server.envelope, "Delivered-To") ||
+		delivered_to_count != ctl->server.envskip)
 		free(line);
 	    else 
 		delivered_to = line;
+	    delivered_to_count++;
 	    continue;
 	}
 
@@ -683,7 +725,7 @@ int readheaders(int sock,
 	    else
 		cp = NULL;
 	    if (cp) {
-		while (*cp && isspace(*cp)) cp++;
+		while (*cp && isspace((unsigned char)*cp)) cp++;
 		if (!*cp || ctl->dropstatus)
 		{
 		    free(line);
@@ -693,7 +735,7 @@ int readheaders(int sock,
 	}
 
 	if (ctl->rewrite)
-	    line = reply_hack(line, ctl->server.truename);
+	    line = reply_hack(line, ctl->server.truename, &linelen);
 
 	/*
 	 * OK, this is messy.  If we're forwarding by SMTP, it's the
@@ -717,7 +759,7 @@ int readheaders(int sock,
 	 * make sure we never try to rewrite such a blank Return-Path.  We
 	 * handle this with a check for <> in the rewrite logic above.
 	 *
-	 * Also, if an email has multiple Return-Path: statement, we only
+	 * Also, if an email has multiple Return-Path: headers, we only
 	 * read the first occurance, as some spam email has more than one
 	 * Return-Path.
 	 *
@@ -865,13 +907,13 @@ int readheaders(int sock,
 
     if (retain_mail)
     {
-	free(msgblk.headers);
-	msgblk.headers = NULL;
 	return(PS_RETAINED);
     }
     if (refuse_mail)
 	return(PS_REFUSED);
     /*
+     * This is the duplicate-message killer code.
+     *
      * When mail delivered to a multidrop mailbox on the server is
      * addressed to multiple people on the client machine, there will
      * be one copy left in the box for each recipient.  This is not a
@@ -883,11 +925,8 @@ int readheaders(int sock,
      * if the mail is addressed to N people, each recipient will
      * get N copies.  This is bad when N > 1.
      *
-     * Foil this by suppressing all but one copy of a message with
-     * a given Message-ID.  The accept_count test ensures that
-     * multiple pieces of email with the same Message-ID, each
-     * with a *single* addressee (the N == 1 case), won't be 
-     * suppressed.
+     * Foil this by suppressing all but one copy of a message with a
+     * given set of headers.
      *
      * Note: This implementation only catches runs of successive
      * messages with the same ID, but that should be good
@@ -899,21 +938,31 @@ int readheaders(int sock,
      * Don't mess with this code casually.  It would be way too easy
      * to break it in a way that blackholed mail.  Better to pass
      * the occasional duplicate than to do that...
+     *
+     * Matthias Andree:
+     * The real fix however is to insist on Delivered-To: or similar
+     * headers and require that one copy per recipient be dropped.
+     * Everything else breaks sooner or later.
      */
-    if (!received_for && env_offs == -1 && !delivered_to)
+    if (MULTIDROP(ctl) && msgblk.headers)
     {
-	if (ctl->lastid && ctl->thisid && !strcasecmp(ctl->lastid, ctl->thisid))
+	MD5_CTX context;
+
+	MD5Init(&context);
+	MD5Update(&context, msgblk.headers, strlen(msgblk.headers));
+	MD5Final(ctl->digest, &context);
+
+	if (!received_for && env_offs == -1 && !delivered_to)
 	{
-	    if (accept_count > 1)
+	    /*
+	     * Hmmm...can MD5 ever yield all zeroes as a hash value?
+	     * If so there is a one in 18-quadrillion chance this 
+	     * code will incorrectly nuke the first message.
+	     */
+	    if (!memcmp(ctl->lastdigest, ctl->digest, DIGESTLEN))
 		return(PS_REFUSED);
 	}
-	else
-	{
-	    if (ctl->lastid)
-		free(ctl->lastid);
-	    ctl->lastid = ctl->thisid;
-	    ctl->thisid = NULL;
-	}
+	memcpy(ctl->lastdigest, ctl->digest, DIGESTLEN);
     }
 
     /*
@@ -927,12 +976,10 @@ int readheaders(int sock,
      */
     if (msgblk.headers == (char *)NULL)
     {
-#ifdef HAVE_SNPRINTF
 	snprintf(buf, sizeof(buf),
-#else
-	sprintf(buf, 
-#endif /* HAVE_SNPRINTF */
-	"From: FETCHMAIL-DAEMON\r\nTo: %s@%s\r\nSubject: Headerless mail from %s's mailbox on %s\r\n",
+		"From: FETCHMAIL-DAEMON\r\n"
+		"To: %s@%s\r\n"
+		"Subject: Headerless mail from %s's mailbox on %s\r\n",
 		user, fetchmailhost, ctl->remotename, ctl->server.truename);
 	msgblk.headers = xstrdup(buf);
     }
@@ -951,7 +998,7 @@ int readheaders(int sock,
 	/* We have the real envelope return-path, stored out of band by
 	 * SDPS - that's more accurate than any header is going to be.
 	 */
-	strcpy(msgblk.return_path, sdps_envfrom);
+	strlcpy(msgblk.return_path, sdps_envfrom, sizeof(msgblk.return_path));
 	free(sdps_envfrom);
     } else
 #endif /* SDPS_ENABLE */
@@ -977,7 +1024,7 @@ int readheaders(int sock,
 	else if (resent_from_offs >= 0 && (ap = nxtaddr(msgblk.headers + resent_from_offs)));
 	else if (from_offs >= 0 && (ap = nxtaddr(msgblk.headers + from_offs)));
 	else if (reply_to_offs >= 0 && (ap = nxtaddr(msgblk.headers + reply_to_offs)));
-	else if (app_from_offs >= 0 && (ap = nxtaddr(msgblk.headers + app_from_offs)));
+	else if (app_from_offs >= 0 && (ap = nxtaddr(msgblk.headers + app_from_offs))) {}
 	/* multi-line MAIL FROM addresses confuse SMTP terribly */
 	if (ap && !strchr(ap, '\n')) {
 	    strncpy(msgblk.return_path, ap, sizeof(msgblk.return_path));
@@ -1007,8 +1054,7 @@ int readheaders(int sock,
       ctl->server.envelope && !strcasecmp(ctl->server.envelope, "Delivered-To"))
    {
 	    find_server_names(delivered_to, ctl, &msgblk.recipients);
-       free(delivered_to);
-       delivered_to = NULL;
+	    xfree(delivered_to);
    }
 	else if (received_for)
 	    /*
@@ -1070,9 +1116,6 @@ int readheaders(int sock,
 	if (outlevel >= O_DEBUG)
 	    report(stdout,
 		   GT_("forwarding and deletion suppressed due to DNS errors\n"));
-	free(msgblk.headers);
-	msgblk.headers = NULL;
-	free_str_list(&msgblk.recipients);
 	return(PS_TRANSIENT);
     }
     else
@@ -1081,9 +1124,6 @@ int readheaders(int sock,
 	if ((n = open_sink(ctl, &msgblk,
 			   &good_addresses, &bad_addresses)) != PS_SUCCESS)
 	{
-	    free(msgblk.headers);
-	    msgblk.headers = NULL;
-	    free_str_list(&msgblk.recipients);
 	    return(n);
 	}
     }
@@ -1111,23 +1151,19 @@ int readheaders(int sock,
     {
 	/* utter any per-message Received information we need here */
         if (ctl->server.trueaddr) {
-#ifdef HAVE_SNPRINTF
+	    char saddr[50];
+	    int e;
+
+	    e = getnameinfo(ctl->server.trueaddr, ctl->server.trueaddr_len,
+		    saddr, sizeof(saddr), NULL, 0,
+		    NI_NUMERICHOST);
+	    if (e)
+		snprintf(saddr, sizeof(saddr), "(%-.*s)", sizeof(saddr) - 3, gai_strerror(e));
 	    snprintf(buf, sizeof(buf),
-#else
-	    sprintf(buf, 
-#endif /* HAVE_SNPRINTF */
-		    "Received: from %s [%u.%u.%u.%u]\r\n", 
-		    ctl->server.truename,
-		    (unsigned char)ctl->server.trueaddr[0],
-		    (unsigned char)ctl->server.trueaddr[1],
-		    (unsigned char)ctl->server.trueaddr[2],
-		    (unsigned char)ctl->server.trueaddr[3]);
+		    "Received: from %s [%s]\r\n", 
+		    ctl->server.truename, saddr);
 	} else {
-#ifdef HAVE_SNPRINTF
-	  snprintf(buf, sizeof(buf),
-#else                       
-	  sprintf(buf,
-#endif /* HAVE_SNPRINTF */
+	    snprintf(buf, sizeof(buf),
 		  "Received: from %s\r\n", ctl->server.truename);
 	}
 	n = stuffline(ctl, buf);
@@ -1137,38 +1173,26 @@ int readheaders(int sock,
 	     * This header is technically invalid under RFC822.
 	     * POP3, IMAP, etc. are not legal mail-parameter values.
 	     */
-#ifdef HAVE_SNPRINTF
 	    snprintf(buf, sizeof(buf),
-#else
-	    sprintf(buf,
-#endif /* HAVE_SNPRINTF */
 		    "\tby %s with %s (fetchmail-%s",
 		    fetchmailhost,
 		    protocol->name,
 		    VERSION);
-	    if (ctl->tracepolls)
+	    if (ctl->server.tracepolls)
 	    {
-		sprintf(buf + strlen(buf), " polling %s account %s",
-			ctl->server.pollname, 
+		snprintf(buf + strlen(buf), sizeof(buf) - strlen(buf),
+			" polling %s account %s",
+			ctl->server.pollname,
 			ctl->remotename);
 	    }
-#ifdef HAVE_SNPRINTF
 	    snprintf(buf+strlen(buf), sizeof(buf)-strlen(buf), ")\r\n");
-#else
-	    strcat(buf, ")\r\n");
-#endif /* HAVE_SNPRINTF */
 	    n = stuffline(ctl, buf);
 	    if (n != -1)
 	    {
 		buf[0] = '\t';
 		if (good_addresses == 0)
 		{
-#ifdef HAVE_SNPRINTF
-		    snprintf(buf+1, sizeof(buf)-1,
-#else
-		    sprintf(buf+1,
-#endif /* HAVE_SNPRINTF */
-			    "for %s (by default); ",
+		    snprintf(buf+1, sizeof(buf)-1, "for <%s> (by default); ",
 			    rcpt_address (ctl, run.postmaster, 0));
 		}
 		else if (good_addresses == 1)
@@ -1176,25 +1200,17 @@ int readheaders(int sock,
 		    for (idp = msgblk.recipients; idp; idp = idp->next)
 			if (idp->val.status.mark == XMIT_ACCEPT)
 			    break;	/* only report first address */
-#ifdef HAVE_SNPRINTF
 		    snprintf(buf+1, sizeof(buf)-1,
-#else                       
-		    sprintf(buf+1,
-#endif /* HAVE_SNPRINTF */
-			    "for %s", rcpt_address (ctl, idp->id, 1));
-		    sprintf(buf+strlen(buf), " (%s); ",
+			    "for <%s>", rcpt_address (ctl, idp->id, 1));
+		    snprintf(buf+strlen(buf), sizeof(buf)-strlen(buf)-1,
+			    " (%s); ",
 			    MULTIDROP(ctl) ? "multi-drop" : "single-drop");
 		}
 		else
 		    buf[1] = '\0';
 
-#ifdef HAVE_SNPRINTF
 		snprintf(buf+strlen(buf), sizeof(buf)-strlen(buf), "%s\r\n",
 			rfc822timestamp());
-#else
-		strcat(buf, rfc822timestamp());
-		strcat(buf, "\r\n");
-#endif /* HAVE_SNPRINTF */
 		n = stuffline(ctl, buf);
 	    }
 	}
@@ -1207,13 +1223,10 @@ int readheaders(int sock,
     {
 	report(stdout, GT_("writing RFC822 msgblk.headers\n"));
 	release_sink(ctl);
-	free(msgblk.headers);
-	msgblk.headers = NULL;
-	free_str_list(&msgblk.recipients);
 	return(PS_IOERR);
     }
-    else if ((run.poll_interval == 0 || nodetach) && outlevel >= O_VERBOSE && !isafile(2))
-	fputs("#", stdout);
+    else if ((run.poll_interval == 0 || nodetach) && outlevel >= O_VERBOSE && !is_a_file(1) && !run.use_syslog)
+	fputc('#', stdout);
 
     /* write error notifications */
     if (no_local_matches || has_nuls || bad_addresses)
@@ -1222,21 +1235,17 @@ int readheaders(int sock,
 	char	errhd[USERNAMELEN + POPBUFSIZE], *errmsg;
 
 	errmsg = errhd;
-	(void) strcpy(errhd, "X-Fetchmail-Warning: ");
+	strlcpy(errhd, "X-Fetchmail-Warning: ", sizeof(errhd));
 	if (no_local_matches)
 	{
 	    if (reject_count != 1)
-		strcat(errhd, GT_("no recipient addresses matched declared local names"));
+		strlcat(errhd, GT_("no recipient addresses matched declared local names"), sizeof(errhd));
 	    else
 	    {
 		for (idp = msgblk.recipients; idp; idp = idp->next)
 		    if (idp->val.status.mark == XMIT_REJECT)
 			break;
-#ifdef HAVE_SNPRINTF
 		snprintf(errhd+strlen(errhd), sizeof(errhd)-strlen(errhd),
-#else
-		sprintf(errhd+strlen(errhd),
-#endif /* HAVE_SNPRINTF */
 			GT_("recipient address %s didn't match any local name"), idp->id);
 	    }
 	}
@@ -1244,34 +1253,24 @@ int readheaders(int sock,
 	if (has_nuls)
 	{
 	    if (errhd[sizeof("X-Fetchmail-Warning: ")])
-#ifdef HAVE_SNPRINTF
 		snprintf(errhd+strlen(errhd), sizeof(errhd)-strlen(errhd), "; ");
 	    snprintf(errhd+strlen(errhd), sizeof(errhd)-strlen(errhd),
-#else
-		strcat(errhd, "; ");
-	    strcat(errhd,
-#endif /* HAVE_SNPRINTF */
 			GT_("message has embedded NULs"));
 	}
 
 	if (bad_addresses)
 	{
 	    if (errhd[sizeof("X-Fetchmail-Warning: ")])
-#ifdef HAVE_SNPRINTF
 		snprintf(errhd+strlen(errhd), sizeof(errhd)-strlen(errhd), "; ");
 	    snprintf(errhd+strlen(errhd), sizeof(errhd)-strlen(errhd),
-#else
-		strcat(errhd, "; ");
-	    strcat(errhd,
-#endif /* HAVE_SNPRINTF */
 			GT_("SMTP listener rejected local recipient addresses: "));
 	    errlen = strlen(errhd);
 	    for (idp = msgblk.recipients; idp; idp = idp->next)
 		if (idp->val.status.mark == XMIT_RCPTBAD)
 		    errlen += strlen(idp->id) + 2;
 
-	    xalloca(errmsg, char *, errlen+3);
-	    (void) strcpy(errmsg, errhd);
+	    errmsg = xmalloc(errlen + 3);
+	    strcpy(errmsg, errhd);
 	    for (idp = msgblk.recipients; idp; idp = idp->next)
 		if (idp->val.status.mark == XMIT_RCPTBAD)
 		{
@@ -1286,6 +1285,9 @@ int readheaders(int sock,
 
 	/* ship out the error line */
 	stuffline(ctl, errmsg);
+
+	if (errmsg != errhd)
+	    free(errmsg);
     }
 
     /* issue the delimiter line */
@@ -1337,7 +1339,7 @@ int readbody(int sock, struct query *ctl, flag forward, int len)
 	    sizeticker += linelen;
 	    while (sizeticker >= SIZETICKER)
 	    {
-		if (outlevel > O_SILENT && (((run.poll_interval == 0 || nodetach) && !isafile(1)) || run.showdots))
+		if (outlevel > O_SILENT && run.showdots && !run.use_syslog)
 		{
 		    fputc('.', stdout);
 		    fflush(stdout);
@@ -1393,7 +1395,7 @@ int readbody(int sock, struct query *ctl, flag forward, int len)
 		release_sink(ctl);
 		return(PS_IOERR);
 	    }
-	    else if (outlevel >= O_VERBOSE && !isafile(1))
+	    else if (outlevel >= O_VERBOSE && !is_a_file(1) && !run.use_syslog)
 	    {
 		fputc('*', stdout);
 		fflush(stdout);
@@ -1407,9 +1409,11 @@ int readbody(int sock, struct query *ctl, flag forward, int len)
 void init_transact(const struct method *proto)
 /* initialize state for the send and receive functions */
 {
+    suppress_tags = FALSE;
     tagnum = 0;
     tag[0] = '\0';	/* nuke any tag hanging out from previous query */
     protocol = (struct method *)proto;
+    shroud[0] = '\0';
 }
 
 static void enshroud(char *buf)
@@ -1443,7 +1447,7 @@ va_dcl
     va_list ap;
 
     if (protocol->tagged && !suppress_tags)
-	(void) sprintf(buf, "%s ", GENSYM);
+        snprintf(buf, sizeof(buf) - 2, "%s ", GENSYM);
     else
 	buf[0] = '\0';
 
@@ -1452,18 +1456,10 @@ va_dcl
 #else
     va_start(ap);
 #endif
-#ifdef HAVE_VSNPRINTF
-    vsnprintf(buf + strlen(buf), sizeof(buf)-strlen(buf), fmt, ap);
-#else
-    vsprintf(buf + strlen(buf), fmt, ap);
-#endif
+    vsnprintf(buf + strlen(buf), sizeof(buf)-2-strlen(buf), fmt, ap);
     va_end(ap);
 
-#ifdef HAVE_SNPRINTF
     snprintf(buf+strlen(buf), sizeof(buf)-strlen(buf), "\r\n");
-#else
-    strcat(buf, "\r\n");
-#endif /* HAVE_SNPRINTF */
     SockWrite(sock, buf, strlen(buf));
 
     if (outlevel >= O_MONITOR)
@@ -1488,7 +1484,7 @@ int size;	/* length of buffer */
     {
 	set_timeout(0);
 	phase = oldphase;
-	if(isidletimeout())
+	if(is_idletimeout())
 	{
 	  resetidletimeout();
 	  return(PS_IDLETIMEOUT);
@@ -1528,7 +1524,7 @@ va_dcl
     phase = SERVER_WAIT;
 
     if (protocol->tagged && !suppress_tags)
-	(void) sprintf(buf, "%s ", GENSYM);
+	snprintf(buf, sizeof(buf) - 2, "%s ", GENSYM);
     else
 	buf[0] = '\0';
 
@@ -1537,19 +1533,14 @@ va_dcl
 #else
     va_start(ap);
 #endif
-#ifdef HAVE_VSNPRINTF
-    vsnprintf(buf + strlen(buf), sizeof(buf)-strlen(buf), fmt, ap);
-#else
-    vsprintf(buf + strlen(buf), fmt, ap);
-#endif
+    vsnprintf(buf + strlen(buf), sizeof(buf)-2-strlen(buf), fmt, ap);
     va_end(ap);
 
-#ifdef HAVE_SNPRINTF
     snprintf(buf+strlen(buf), sizeof(buf)-strlen(buf), "\r\n");
-#else
-    strcat(buf, "\r\n");
-#endif /* HAVE_SNPRINTF */
-    SockWrite(sock, buf, strlen(buf));
+    if (SockWrite(sock, buf, strlen(buf)) < strlen(buf)) {
+	/* short write, bail out */
+	return PS_SOCKET;
+    }
 
     if (outlevel >= O_MONITOR)
     {
